#Brian Carty
#PhD Code
#MultiBin, August 2010

import csnd
from Tkinter import *
import tkSimpleDialog
import tkMessageBox
import cmath, math

#inherit frame...class is a frame!
class Application(Frame):

    #global reverb on/off
    reverb = 0
    #how many sources currently active?
    count = 0
    #how many sources removed in current scene
    removed = 0
    #how many sources have been 'clear alled'
    allremoved = 0
    #source instrument on?
    playing = 0
      
    def move(self, event):
        #widget that called the event
        loc = event.widget
        #set, in case of canvas scroll?...
        x = loc.canvasx(event.x)
        y = loc.canvasy(event.y)
        #find_withtag returns list (tuple) of matching items, in order created:
        #only 1 item will be returned here (or if they are at same loc, most recently created)...
        item = loc.find_withtag("current")      
        #text only has 2 coords...
        loc.coords(item, x, y)
        #choose value based on default room sizes/inputted size
        #head is item 2, oval 3, rectangle 1, then sources in order created...
        #if > no of items active, subtract no of removed...also subtract total removed...
        if item[0] - 3 - self.allremoved > self.count:
            chno = item[0] - 3 - self.allremoved - self.removed
        else:
            chno = item[0] - 3 - self.allremoved

        self.cs.SetChannel("xsrc%d" %chno, (x / self.sizex) * self.roomarray[0])
        #send inverted y to csound...
        y = self.sizey - y
        self.cs.SetChannel("ysrc%d" %chno, (y / self.sizey) * self.roomarray[1])

    #a simpler move function, as head will always be on same channel
    def movehead(self, event):
        #widget that called the event
        loc = event.widget
        #set, in case of canvas scroll?...
        self.headx = loc.canvasx(event.x)
        #invert y throughout
        self.heady = loc.canvasy(event.y)
        #y = self.sizey - y
        item = loc.find_withtag("current")[0]       
        #head coords...include existing measured rotation...
        loc.coords(item, self.headx + self.rot[0].real, self.heady + self.rot[0].imag, 
                   self.headx + self.rot[1].real, self.heady + self.rot[1].imag, 
                   self.headx + self.rot[2].real, self.heady + self.rot[2].imag, 
                   self.headx + self.rot[3].real, self.heady + self.rot[3].imag, 
                   self.headx + self.rot[4].real, self.heady + self.rot[4].imag, 
                   self.headx + self.rot[5].real, self.heady + self.rot[5].imag, 
                   self.headx + self.rot[6].real, self.heady + self.rot[6].imag)
        #move range 
        self.canvas.coords(self.oval, self.headx - self.range, self.heady - self.range, 
                           self.headx + self.range, self.heady + self.range)
        self.cs.SetChannel("xhead", (self.headx / self.sizex) * self.roomarray[0])
        #send inverted y to csound...
        y = self.sizey - self.heady
        self.cs.SetChannel("yhead", (y / self.sizey) * self.roomarray[1])

    def select(self, event):
        loc = event.widget
        item = loc.find_withtag("current")
        #red if selected!
        loc.itemconfig(item, fill = "red")

    def deselect(self, event):
        loc = event.widget
        item = loc.find_withtag("current")
        loc.itemconfig(item, fill = "blue")

    def play(self):
        self.perf.InputMessage("i1 0 -1")
        self.playing = 1
    
    def stop(self):
        if self.playing:
            self.perf.InputMessage("i-1 0 -1")
            self.playing = 0

    def headrotate(self, event):
        degrees = self.rotscale.get()
        self.cs.SetChannel("rot", degrees)
        #rotate polygon...use complex maths here, as it more elegant
        offset = complex(self.headx, self.heady)
        radangle = math.radians(degrees)
        compangle = cmath.exp(radangle * 1j)
        #angle, from centre of non rotated polygon, rotate by point at 0, 0, add offset again at end...        
        self.rot[0] = compangle * (complex(self.headx + 5, self.heady + 10) - offset)
        self.rot[1] = compangle * (complex(self.headx + 10, self.heady) - offset)
        self.rot[2] = compangle * (complex(self.headx + 5, self.heady - 10) - offset)
        self.rot[3] = compangle * (complex(self.headx, self.heady - 13) - offset)
        self.rot[4] = compangle * (complex(self.headx - 5, self.heady - 10) - offset) 
        self.rot[5] = compangle * (complex(self.headx - 10, self.heady) - offset) 
        self.rot[6] = compangle * (complex(self.headx - 5, self.heady + 10) - offset)
        #add offset again...
        self.canvas.coords(self.head, self.rot[0].real + self.headx, self.rot[0].imag + self.heady, 
                           self.rot[1].real + self.headx, self.rot[1].imag + self.heady, 
                           self.rot[2].real + self.headx, self.rot[2].imag + self.heady, 
                           self.rot[3].real + self.headx, self.rot[3].imag + self.heady, 
                           self.rot[4].real + self.headx, self.rot[4].imag + self.heady, 
                           self.rot[5].real + self.headx, self.rot[5].imag + self.heady, 
                           self.rot[6].real + self.headx, self.rot[6].imag + self.heady)

    def lateamp(self, event):
        vol = self.latescale.get()
        self.cs.SetChannel("lateamp", vol)

    def end(self):
        self.perf.Stop()
        self.perf.Join()
        self.master.destroy()

    #generic: new speaker
    def newspeaker(self, x, y):
        self.count = self.count + 1
        no = str(self.count)

        self.canvas.create_text(x, y, text = no, fill = 'blue', tags = 'drag')

        #invert y for csound
        y = self.sizey - y
        
        S = 'i101.{0} 0 -1 {0} {1} {2} {3} {4} {5} {6} {7} {8} {9} {10} \
            {11} {12} {13} {14} {15} {16} {17} {18} {19} {20}' \
            .format(self.count, (x / self.sizex) * self.roomarray[0], (y / self.sizey) * self.roomarray[1], \
            self.roomarray[0], self.roomarray[1], self.roomarray[2], self.roomarray[3], self.roomarray[4], \
            self.roomarray[5], self.roomarray[6], self.roomarray[7], self.roomarray[8], self.roomarray[9], \
            self.roomarray[10], self.roomarray[11], self.roomarray[12], self.roomarray[13], self.roomarray[14], \
            self.roomarray[15], self.roomarray[16], self.roomarray[17])

        #set channel
        self.cs.SetChannel("xsrc%d" %self.count, (x / self.sizex) * self.roomarray[0])
        self.cs.SetChannel("ysrc%d" %self.count, (y / self.sizey) * self.roomarray[1])
        self.perf.InputMessage(S)

        #turn on global late reverb, if not already on (may have added and removed sources!)
        if self.count == 1:
            if self.reverb == 0:
                self.perf.InputMessage("i102 0 -1")
                self.reverb = 1
                print "reverb"

    def newsrc(self):
        #default add source: in front of listener
        self.src = location(self)
        if self.src.flag:
            mode = self.src.locationdata[0]
            if mode == 1:
                self.newpolar(self.src.locationdata[2], self.src.locationdata[1])
            #simpler scenario...
            elif mode == 2:
                self.srcx = self.src.locationdata[1]
                self.srcy = self.src.locationdata[2]
        
            #invert y for pixel location
            self.srcy = self.sizey - self.srcy
            #validate location, leave other validation up to csound...wall params...room limits...
            #src must be minimum 10 cm from wall, as per csound
            if self.srcx > self.sizex - (.1 / self.roomarray[0]) * self.sizex:
                self.srcx = self.sizex - (.1 / self.roomarray[0]) * self.sizex
            if self.srcx < (.1 / self.roomarray[0]) * self.sizex:
                self.srcx = (.1 / self.roomarray[0]) * self.sizex
            if self.srcy > self.sizey - (.1 / self.roomarray[1]) * self.sizey:
                self.srcy = self.sizey - (.1 / self.roomarray[1]) * self.sizey
            if self.srcy < (.1 / self.roomarray[1]) * self.sizey:
                self.srcy = (.1 / self.roomarray[1]) * self.sizey
                
            self.newspeaker(self.srcx, self.srcy)

    def newpolar(self, angle, distance):
        radangle = math.radians(angle)
        #angle measured from centre, clockwise...
        #sin gives x coord...cos y...
        #srcx and srcy are temp variables used for each source
        self.srcx = math.sin(radangle)
        self.srcy = math.cos(radangle)
        #radius: same if calculated from x or y params, as per setup
        mult = self.sizex / self.roomarray[0] * distance
        self.srcx *= mult
        self.srcy *= mult
        #relative to listener/centre
        self.srcx += self.sizex / 2
        self.srcy += self.sizey / 2	

    #each default setup will call a unique function...
    #need different menu for different possible ambi, vbap setups in csound...
    #mode 4 ambisonics
    def ambi4(self):
        self.clearall()
        #local variables here
        #angles: anticlockwise
        ang = -22.5
        if self.roomarray[0] < self.roomarray[1]:
            dist = self.roomarray[0] / 3
        else:
            dist = self.roomarray[1] / 3
        for i in range(1, 9):
            self.newpolar(ang, dist)
            #invert y for pixel location
            self.srcy = self.sizey - self.srcy
            self.newspeaker(self.srcx, self.srcy)
            ang -= 45

    #from manual example...
    def vbap8(self):
        self.clearall()
        ang = 15
        if self.roomarray[0] < self.roomarray[1]:
			dist = self.roomarray[0] / 3
        else:
            dist = self.roomarray[1] / 3
        for i in range(1, 9):
            self.newpolar(ang, dist)
            #invert y for pixel location
            self.srcy = self.sizey - self.srcy
            self.newspeaker(self.srcx, self.srcy)
            if i == 4:
                ang += 30
            else:
                ang += 50
	
	#simple, externalised stereo
    def stereo(self):
		self.clearall()
		if self.roomarray[0] < self.roomarray[1]:
			dist = self.roomarray[0] / 3
		else:
		    dist = self.roomarray[1] / 3
		self.newpolar(30, dist)
		#invert y for pixel location
		self.srcy = self.sizey - self.srcy
		self.newspeaker(self.srcx, self.srcy)
		self.newpolar(-30, dist)
		self.srcy = self.sizey - self.srcy
		self.newspeaker(self.srcx, self.srcy)

    def clear(self):
        #last element of array = [-1]
        recent = self.canvas.find_all()[-1]
        self.canvas.delete(recent)
        self.perf.InputMessage("i-101.%d 0 -1" %self.count)
        print "i-101.%d 0 -1 SENT" %self.count

    def clearrecent(self):
        if self.count > 0:
            self.clear()
            self.count -= 1
            self.removed += 1

    def clearall(self): 
        self.stop()
        while self.count > 0:
            self.clear()
            self.count -= 1
            self.allremoved += 1
        self.allremoved += self.removed
        self.removed = 0

    #get parameters...
    def newscene(self):
    #stop playback
        self.stop()
        self.rm = room(self)
        #if not cancel...
        if self.rm.flag:
            self.clearall()
            #turn off reverb
            if self.reverb == 1:
                self.perf.InputMessage("i-102 0 -1")
                self.reverb = 0
                print "reverb off"
            #globals off
            self.perf.InputMessage("i-100 0 -1")

            for i in range(18):
                self.roomarray[i] = self.rm.rmdata[i]          

            self.sizey = self.sizex / self.roomarray[0] * self.roomarray[1]
            self.canvas.configure(width = self.sizex, height = self.sizey)
            #useful canvas area...needs to be min .1m from wall
            self.canvas.coords(self.rect, (.1 / self.roomarray[0]) * self.sizex + 1, (.1 / self.roomarray[1]) * self.sizey + 1, 
                               self.sizex + 2 - (.1 / self.roomarray[0]) * self.sizex, 
                               self.sizey + 2 - (.1 / self.roomarray[1]) * self.sizey)   
            self.canvas.itemconfigure(self.rect, fill = "white", outline = "grey")
            #initialise
            self.headx = self.sizex / 2
            self.heady = self.sizey / 2
            #reset rotation
            self.rotscale.set(0)
            #reset amp
            self.latescale.set(0.3)
            #move 'head' to middle
            self.canvas.coords(self.head, self.headx + 5, self.heady + 10, 
                               self.headx + 10, self.heady,
                               self.headx + 5, self.heady - 10, 
                               self.headx, self.heady - 13, 
                               self.headx - 5, self.heady - 10, 
                               self.headx - 10, self.heady, 
                               self.headx - 5, self.heady + 10)
            #near field range
            self.range = self.sizex / self.roomarray[0] * .45
            self.canvas.coords(self.oval, self.headx - self.range, self.heady - self.range, 
                               self.headx + self.range, self.heady + self.range)   
            #head back to middle!
            self.cs.SetChannel("xhead", self.roomarray[0] / 2)
            self.cs.SetChannel("yhead", self.roomarray[1] / 2)
            #rotation back to 0
            self.cs.SetChannel("rot", 0)
            self.statusstring = ("x: %.2f, y: %.2f z: %.2f" %(self.roomarray[0], self.roomarray[1], self.roomarray[2]))
            self.status.config(text = self.statusstring)
            self.status.update_idletasks()
            # globals back on
            self.perf.InputMessage("i100 0 -1 %f %f" %(self.roomarray[0] / 2, self.roomarray[1] / 2))
            
    def about(self):
        tkMessageBox.showwarning("Hi", "MultiBin v 1.0\nAugust 2010\nby Brian Carty\nUses Csound opcodes hrtfealry & hrtfreverb\nfor help and blurb, see bmcarty.com")

    def __init__(self, master = None):
        master.title("MultiBin")
        
        Frame.__init__(self, master)

        #csound setup: turn on, wait for input!
        self.cs = csnd.Csound()
        self.cs.Compile("MultiBin.csd")
        self.perf = csnd.CsoundPerformanceThread(self.cs)
        self.perf.Play()

        #create a Menu base
        self.menu = Menu(self)
        #add it
        self.master.config(menu = self.menu)
        #create menu
        self.filemenu = Menu(self.menu)
        #file menu
        self.menu.add_cascade(label = "File", menu = self.filemenu)
        #this choice is required before processing begins!
        self.filemenu.add_command(label = "New Scene(==Restart)", command = self.newscene)
        self.filemenu.add_command(label = "New Source", command = self.newsrc)
        self.filemenu.add_separator()
        self.filemenu.add_command(label = "'Ideal' Stereo", command = self.stereo)
        #ambisonics: layout 4 from bformdec1
        self.filemenu.add_command(label = "Ambi - Octogon", command = self.ambi4)
        self.filemenu.add_command(label = "VBAP - 8 Channel", command = self.vbap8)
        self.filemenu.add_separator()
        #clear
        self.filemenu.add_command(label = "Clear Recent", command = self.clearrecent)
        self.filemenu.add_command(label = "Clear All", command = self.clearall)
        self.filemenu.add_separator()
        self.filemenu.add_command(label = "Exit", command = self.end)
        self.helpmenu = Menu(self.menu)
        self.menu.add_cascade(label = "Help", menu = self.helpmenu)
        self.helpmenu.add_command(label = "About...", command = self.about)

        #button for source on/off, dial for head rotation
        self.playbut = Button(master, text = "Start Playback Instr", command = self.play)
        self.stopbut = Button(master, text = "Stop", command = self.stop)

        #scale widget for head rotation
        self.rotscale = Scale(master, orient = HORIZONTAL, label = "Head Rotation", 
                              from_ = -90.0, to = 90.0, length = 120, cursor = "exchange", 
                              resolution = .1, command = self.headrotate)

        #scale for output level of late reverb
        self.latescale = Scale(master, orient = HORIZONTAL, label = "Late Amp", 
                              from_ = 0, to = .99, length = 120, 
                              resolution = .01, command = self.lateamp)

        #status: hints for user
        self.statusstring = "default room: x: 10.00, y: 10.00, z: 3.00"
        self.status = Label(master, text = self.statusstring, relief = SUNKEN, anchor = W)

        #canvas, default size for first run...
        self.canvas = Canvas(master, bg = "grey", width = 400, height = 400)

        #setup grid...
        self.playbut.grid(row = 0, column = 0)
        self.stopbut.grid(row = 0, column = 1)
        self.rotscale.grid(row = 0, column = 2)
        self.latescale.grid(row = 0, column = 3)
        self.canvas.grid(row = 1, column = 0, columnspan = 4)
        self.status.grid(row = 2, column = 0, columnspan = 4, sticky = E + W)

        #set reverb...
        self.latescale.set(0.3)
        
        #defaults
        #roomarray data order: rmx, rmy, rmz, wlh, wll, wl1, wl2, wl3, flh, fll, fl1, fl2, fl3, clh, cll, cl1, cl2, cl3
        self.roomarray = [10.0, 10.0, 10.0, 3.0, .3, .1, .75, .95, .9, .6, .1, .95, .6, .35, .2, .1, 1.0, 1.0, 1.0]

        self.sizex = 400.0
        self.sizey = self.sizex / self.roomarray[0] * self.roomarray[1]
        #useful canvas area...needs to be min .1m from wall
        #canvas goes from 1 - 401, add extra 1 to bottom right corner as rect is contained within this point
        self.rect = self.canvas.create_rectangle((.1 / self.roomarray[0]) * self.sizex + 1, (.1 / self.roomarray[1]) * self.sizey + 1, 
                                                 self.sizex + 2 - (.1 / self.roomarray[0]) * self.sizex, 
                                                 self.sizey + 2 - (.1 / self.roomarray[1]) * self.sizey, 
                                                 fill = "white", outline = "grey")
        #initialise
        self.headx = self.sizex / 2
        self.heady = self.sizey / 2
        #initialise to zero...for rotation
        self.rot = []
        for i in range (7):
            self.rot.append(complex(0, 0))
        #head best as last object, most recent will be selected first!
        self.range = self.sizex / self.roomarray[0] * .45
        self.oval = self.canvas.create_oval(self.headx - self.range, self.heady - self.range, 
                                            self.headx + self.range, self.heady + self.range, outline = "grey")
        #points for polygon of head...
        self.head = self.canvas.create_polygon(self.headx + 5, self.heady + 10, 
                                               self.headx + 10, self.heady, 
                                               self.headx + 5, self.heady - 10, 
                                               self.headx, self.heady - 13, 
                                               self.headx - 5, self.heady - 10, 
                                               self.headx - 10, self.heady, 
                                               self.headx - 5, self.heady + 10, 
                                               fill = "green", tags = "draghead", outline = "black")
        self.cs.SetChannel("xhead", self.roomarray[0] / 2)
        self.cs.SetChannel("yhead", self.roomarray[1] / 2)
        self.cs.SetChannel("rot", 0)
        self.cs.SetChannel("lateamp", 0.3)
        self.perf.InputMessage("i100 0 -1 %f %f" %(self.roomarray[0] / 2, self.roomarray[1] / 2))
        #link 'drag' to functions
        self.canvas.tag_bind('drag','<B1-Motion>', self.move)
        self.canvas.tag_bind('drag','<ButtonPress>', self.select)
        self.canvas.tag_bind('drag','<ButtonRelease>', self.deselect)
        self.canvas.tag_bind('draghead','<B1-Motion>', self.movehead)
        #closing window also ends csd...
        self.master.protocol("WM_DELETE_WINDOW", self.end)

#dialog for room creation
class room(tkSimpleDialog.Dialog):

    flag = 0
    
    def body(self, master):
        self.complex = IntVar()
        Checkbutton(master, variable = self.complex, text = "Show Complex Params?", command = self.check).grid(row = 0, column = 0)
        Label(master, text = "Room X:").grid(row = 2, column = 0)
        Label(master, text = "Room Y:").grid(row = 3, column = 0)
        Label(master, text = "Room Z:").grid(row = 4, column = 0)

        #empty list
        self.r = []
        for i in range (3):
            self.r.append(Entry(master, width = 10))   
        self.r[0].insert(0, "10")
        self.r[1].insert(0, "10")
        self.r[2].insert(0, "3")
        for i in range (3):
            self.r[i].grid(row = i + 2, column = 1)
        Label(master, text = "High Ab Coef:").grid(row = 1, column = 2)
        Label(master, text = "Low Ab Coef:").grid(row = 2, column = 2)
        Label(master, text = "Band 1(cf: 250Hz):").grid(row = 3, column = 2)
        Label(master, text = "Band 2(cf: 1000Hz):").grid(row = 4, column = 2)
        Label(master, text = "Band 3(cf: 4000Hz):").grid(row = 5, column = 2)
        Label(master, text = "Walls").grid(row = 0, column = 3)

        self.w = []
        for i in range (5):
            self.w.append(Entry(master, width = 10))
        self.w[0].insert(0, ".3")
        self.w[1].insert(0, ".1")
        self.w[2].insert(0, ".75")
        self.w[3].insert(0, ".95")
        self.w[4].insert(0, ".9")
        for i in range (5):
            self.w[i].grid(row = i + 1, column = 3)
       
        Label(master, text = "Floor").grid(row = 0, column = 4)
        self.f = []
        for i in range (5):
            self.f.append(Entry(master, width = 10))
        self.f[0].insert(0, ".6")
        self.f[1].insert(0, ".1")
        self.f[2].insert(0, ".95")
        self.f[3].insert(0, ".6")
        self.f[4].insert(0, ".35")
        for i in range (5):
            self.f[i].grid(row = i + 1, column = 4)
    
        Label(master, text = "Ceiling").grid(row = 0, column = 5)
        self.c = []
        for i in range (5):
            self.c.append(Entry(master, width = 10))
        self.c[0].insert(0, ".2")
        self.c[1].insert(0, ".1")
        self.c[2].insert(0, "1.0")
        self.c[3].insert(0, "1.0")
        self.c[4].insert(0, "1.0")
        for i in range (5):
            self.c[i].grid(row = i + 1, column = 5)

        #disable extra parameters by default
        for i in range (5):
            self.w[i].configure(state = DISABLED)
            self.f[i].configure(state = DISABLED)
            self.c[i].configure(state = DISABLED)
            
        #initial focus  
        return self.r[0] 

    def apply(self):
        #avoid error on cancel with flag...
        self.flag = 1
        self.rmdata = []
        for i in range (3):
            self.rmdata.append(float(self.r[i].get()))
        for i in range (5):
            self.rmdata.append(float(self.w[i].get()))
        for i in range (5):
            self.rmdata.append(float(self.f[i].get()))
        for i in range (5):
            self.rmdata.append(float(self.c[i].get()))
            
    def check(self):
        if self.complex.get() == 0:
            for i in range (5):
                self.w[i].configure(state = DISABLED)
                self.f[i].configure(state = DISABLED)
                self.c[i].configure(state = DISABLED)
        else:
            for i in range (5):
                self.w[i].configure(state = NORMAL)
                self.f[i].configure(state = NORMAL)
                self.c[i].configure(state = NORMAL)

class location(tkSimpleDialog.Dialog):

    flag = 0

    def body(self, master):

        self.mode = IntVar()
        self.r1 = Radiobutton(master, text = "Angle, Distance Input", value = 1, variable = self.mode, command = self.rect)
        #need to do this separately to return correct type self.r1...
        self.r1.grid(row = 0, column = 0, columnspan = 2)
        Label(master, text="OR").grid(row = 0, column = 2)
        Radiobutton(master, text = "X, Y Coordinates", value = 2, variable = self.mode, command = self.polar).grid(row = 0, column = 3, columnspan = 2)
        Label(master, text="Note: All values must fit on canvas/in room specified, and will be truncated accordingly!").grid(row = 1, columnspan = 4)
        Label(master, text="Distance from Centre:").grid(row = 2, column = 0)
        Label(master, text="Angle:").grid(row = 3, column = 0)
        Label(master, text="X:").grid(row = 2, column = 3)
        Label(master, text="Y:").grid(row = 3, column = 3)
        self.r1.select()

        self.e = []
        for i in range (4):
            self.e.append(Entry(master, width = 10))   
        self.e[0].insert(0, "1")
        self.e[1].insert(0, "0")
        self.e[2].insert(0, "100")
        self.e[3].insert(0, "100")
        for i in range (2):
            self.e[i].grid(row = i + 2, column = 1)
        for i in range (2):
            self.e[i + 2].grid(row = i + 2, column = 4) 

        for i in range (2):
            self.e[i + 2].configure(state = DISABLED) 

        #initial focus
        return self.e[0] 

    def apply(self):
        self.flag = 1
        dist = float(self.e[0].get())
        angle = float(self.e[1].get())
        x = float(self.e[2].get())
        y = float(self.e[3].get())
        temp = self.mode.get()
        #fill in array...
        if temp == 1:
            self.locationdata = 1, dist, angle
        elif temp == 2:
            self.locationdata = 2, x, y

    def rect(self):
        for i in range (2):
            self.e[i].configure(state = NORMAL)
        for i in range (2):
            self.e[i+ 2].configure(state = DISABLED) 

    def polar(self):
        for i in range (2):
            self.e[i].configure(state = DISABLED)
        for i in range (2):
            self.e[i+ 2].configure(state = NORMAL)
        
app = Application(Tk())       
app.mainloop()

