#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include "csound.h" //needed for cscore
#include "cscore.h"
#include "math.h"

int *pitchinputarray; //global to pass values to pitch functions
float *rhyinputarray; //global pass values to rhythmic functions

char path[] = "/opt/csound513/Js/";  //!!IMPORTANT: set your path here

int main(void)
{

    int numpitches, numrhys;
    int *pitchreturnarray; //values returned from pitch functions
    float *rhyreturnarray; //values returned from rhythmic functions
    float *p3array;
    int *p4array;

    CSOUND *cs; //c version, csound.h. //score variables
    // cs = new CSOUND;
    FILE    *insco, *outsco;
    int     result;

    const char *csvector[] = {"testin.sco", "testout.sco"};
    //uses a different vector from argv so that argv can be
    //used below to run csound
    int argc = 2;

//**************************************
//********  GENERATE PITCHES ***********
//**************************************

    int pitchdist = 1;

    printf("Enter a distribution for pitches, \
            \n1 = fractal, 2 = arcsine, 3 = beta: ");
    scanf("%d",&pitchdist);

    switch(pitchdist)
    {
    case 1:
        printf("you chose fractal\n");
        printf ("\n");
        pitchreturnarray = fractalinput(pitchinputarray);
        break;
    case 2:
        printf("you chose arcsine\n");
        printf ("\n");
        pitchreturnarray = arcsineinput(pitchinputarray);
        break;
    case 3:
        printf("you chose beta\n");
        printf ("\n");
        pitchreturnarray = betainput(pitchinputarray);
        break;
    default:
        printf("you did not choose a pitch distribution\n");
    }

//**************************************
//********  PRINT PITCHES **************
//**************************************
    printf("\n...from main.c\n");

    numpitches = 1;
    do
    {
        printf("%4d",pitchreturnarray[numpitches]);/*prints numbers */
        numpitches++;
    }
    while (pitchreturnarray[numpitches]>0);
    printf("\n\n");

//**************************************
//********  GENERATE RHYTHM ************
//**************************************
    int dims = 1;

    printf("Enter a multidimensional array for rhythm, \
            \n(1 = 1 dim, 2 = 2 dim, 3 = 3 dim, 4 = 4 dim): ");
    scanf("%d",&dims);

    switch(dims)
    {
    case 1:
        printf("you chose one dim\n");
        printf ("\n");
        rhyreturnarray = rhy1Dinput(rhyinputarray);
        break;
    case 2:
        printf("you chose two dims\n");
        printf ("\n");
        rhyreturnarray = rhy2Dinput(rhyinputarray);
        break;
    case 3:
        printf("you chose three dims\n");
        printf ("\n");
        rhyreturnarray = rhy3Dinput(rhyinputarray);
        break;
    case 4:
        printf("you chose four dims\n");
        printf ("\n");
        rhyreturnarray = rhy4Dinput(rhyinputarray);
        break;
    default:
        printf("you did not choose any dims\n");
    }
//**************************************
//********  PRINT RHYTHMS **************
//**************************************

    printf("\n...from main.c\n");
    numrhys = 1;  //note problem passing to cscore2
    do
    {
        printf("%4.6f\n",rhyreturnarray[numrhys]);/*prints numbers */
        numrhys++;
        //} while (arrayindex<numrhys+1);
    }
    while (rhyreturnarray[numrhys]>0.00001); //problem with floating pt.
                                             //cannot just use >0
    printf("\n\n");

//**************************************
//********  CSCORE (from frontends) ****
//**************************************

    /* Standalone Cscore is now a client of the Csound API */

    result = csoundInitialize(&argc, (char ***)csvector, 0);
    if (result < 0)
    {
        fprintf(stderr, "Could not initialize the Csound library.\n");
        return -1;
    }
    cs = csoundCreate(NULL);
    if (cs == NULL)
    {
        fprintf(stderr, "Could not instantiate Csound.\n");
        return -1;
    }

    if (argc < 2)
    {
        fprintf(stderr, "Insufficient arguments: must provide an input filename.\n");
        return -1;
    }
    if (!(insco = fopen(csvector[0], "r")))
    {
        fprintf(stderr, "Cannot open the input score '%s'\n", csvector[0]);
        return -1;
    }


    if (argc < 3)
    {
        if (!(outsco = fopen(csvector[1], "w")))
        {
            fprintf(stderr, "Cannot open the output score '%s'\n", csvector[1]);
            return -1;
        }
    }

    csoundInitializeCscore(cs, insco, outsco);
    cscore2(cs, rhyreturnarray, pitchreturnarray, numpitches, numrhys);  // call the user cscore program
    // parameters include rhys, pitches, and their respective sizes

    fclose(insco); //may not be needed
    fclose(outsco); //may not be needed

//**************************************
//********  CSOUND PLAYER **************
//**************************************
    //this a csound API PerformKsmps loop to play the sounds
    //instead of using the preprocessor's play function.
    //It may or may not be working properly.

    //reinitialize these values for use with csound compiler
    //from cscore use above
    /*
        //clear the argv array used above for cscore
    //    memset( argv, 0, sizeof(argv) );  //use memory.h
    const char *argv[] = {"csound", "testout.orc", "testout.sco"};
        argc = 3;
        //uses also .csoundrc for additional command line parameters
        //be sure to set the environment variable path for CSOUNDRC

     //initialise Csound library
        int myinit = csoundInitialize(&argc, (char ***)argv, 0);
        //Debug:printf("\n ***myinit result = %d", myinit);

        //Creates an instance of Csound.
        cs = csoundCreate(NULL);

        //compile and test the result
        result= csoundCompile(cs, 3, (char **)argv);
        //Debug:printf("*****compile result = %d", result);

        //int perftest = csoundPerformKsmps(csound);
        //Debug:printf("\n *******perftest = %d\n", perftest);

    	if(!result)
    	{
    	 while(csoundPerformKsmps(cs)==0){
           }

    	}
        csoundDestroy(cs);


        return result;
    //=================================//end play .csd
    */


    return 0;



}
